@@ -23,11 +23,11 @@ from typing import Set
 from typing import Tuple
 from typing import Union
 
-import zmq  # type: ignore
-from jupyter_core.paths import jupyter_data_dir  # type: ignore
+import zmq
+from jupyter_core.paths import jupyter_data_dir
 from jupyter_core.paths import jupyter_runtime_dir
 from jupyter_core.paths import secure_write
-from traitlets import Bool  # type: ignore
+from traitlets import Bool
 from traitlets import CaselessStrEnum
 from traitlets import Instance
 from traitlets import Int
@@ -35,7 +35,7 @@ from traitlets import Integer
 from traitlets import observe
 from traitlets import Type
 from traitlets import Unicode
-from traitlets.config import LoggingConfigurable  # type: ignore
+from traitlets.config import LoggingConfigurable
 from traitlets.config import SingletonConfigurable
 
 from .localinterfaces import localhost
@@ -117,7 +117,7 @@ def write_connection_file(
         + int(hb_port <= 0)
     )
     if transport == "tcp":
-        for i in range(ports_needed):
+        for _ in range(ports_needed):
             sock = socket.socket()
             # struct.pack('ii', (0,0)) is 8 null bytes
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, b"\0" * 8)
@@ -129,7 +129,7 @@ def write_connection_file(
             ports.append(port)
     else:
         N = 1
-        for i in range(ports_needed):
+        for _ in range(ports_needed):
             while os.path.exists("%s-%s" % (ip, str(N))):
                 N += 1
             ports.append(N)
@@ -160,37 +160,25 @@ def write_connection_file(
 
     # Only ever write this file as user read/writeable
     # This would otherwise introduce a vulnerability as a file has secrets
-    # which would let others execute arbitrarily code as you
+    # which would let others execute arbitrary code as you
     with secure_write(fname) as f:
         f.write(json.dumps(cfg, indent=2))
 
     if hasattr(stat, "S_ISVTX"):
-        # set the sticky bit on the file and its parent directory
-        # to avoid periodic cleanup
-        paths = [fname]
+        # set the sticky bit on the parent directory of the file
+        # to ensure only owner can remove it
         runtime_dir = os.path.dirname(fname)
         if runtime_dir:
-            paths.append(runtime_dir)
-        for path in paths:
-            permissions = os.stat(path).st_mode
+            permissions = os.stat(runtime_dir).st_mode
             new_permissions = permissions | stat.S_ISVTX
             if new_permissions != permissions:
                 try:
-                    os.chmod(path, new_permissions)
+                    os.chmod(runtime_dir, new_permissions)
                 except OSError as e:
-                    if e.errno == errno.EPERM and path == runtime_dir:
+                    if e.errno == errno.EPERM:
                         # suppress permission errors setting sticky bit on runtime_dir,
                         # which we may not own.
                         pass
-                    else:
-                        # failed to set sticky bit, probably not a big deal
-                        warnings.warn(
-                            "Failed to set sticky bit on %r: %s"
-                            "\nProbably not a big deal, but runtime files may be cleaned up "
-                            "periodically." % (path, e),
-                            RuntimeWarning,
-                        )
-
     return fname, cfg
 
 
@@ -448,9 +436,8 @@ class ConnectionFileMixin(LoggingConfigurable):
     def blocking_client(self):
         """Make a blocking client connected to my kernel"""
         info = self.get_connection_info()
-        info["parent"] = self
-        bc = self.blocking_class(**info)
-        bc.session.key = self.session.key
+        bc = self.blocking_class(parent=self)
+        bc.load_connection_info(info)
         return bc
 
     def cleanup_connection_file(self) -> None:
@@ -463,7 +450,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             self._connection_file_written = False
             try:
                 os.remove(self.connection_file)
-            except (IOError, OSError, AttributeError):
+            except (OSError, AttributeError):
                 pass
 
     def cleanup_ipc_files(self) -> None:
@@ -474,7 +461,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             ipcfile = "%s-%i" % (self.ip, port)
             try:
                 os.remove(ipcfile)
-            except (IOError, OSError):
+            except OSError:
                 pass
 
     def _record_random_port_names(self) -> None:
@@ -580,18 +567,70 @@ class ConnectionFileMixin(LoggingConfigurable):
         if "signature_scheme" in info:
             self.session.signature_scheme = info["signature_scheme"]
 
-    def _force_connection_info(self, info: KernelConnectionInfo) -> None:
-        """Unconditionally loads connection info from a dict containing connection info.
+    def _reconcile_connection_info(self, info: KernelConnectionInfo) -> None:
+        """Reconciles the connection information returned from the Provisioner.
 
-        Overwrites connection info-based attributes, regardless of their current values
-        and writes this information to the connection file.
+        Because some provisioners (like derivations of LocalProvisioner) may have already
+        written the connection file, this method needs to ensure that, if the connection
+        file exists, its contents match that of what was returned by the provisioner.  If
+        the file does exist and its contents do not match, a ValueError is raised.
+
+        If the file does not exist, the connection information in 'info' is loaded into the
+        KernelManager and written to the file.
         """
-        # Reset current ports to 0 and indicate file has not been written to enable override
-        self._connection_file_written = False
-        for name in port_names:
-            setattr(self, name, 0)
-        self.load_connection_info(info)
-        self.write_connection_file()
+        # Prevent over-writing a file that has already been written with the same
+        # info.  This is to prevent a race condition where the process has
+        # already been launched but has not yet read the connection file - as is
+        # the case with LocalProvisioners.
+        file_exists: bool = False
+        if os.path.exists(self.connection_file):
+            with open(self.connection_file) as f:
+                file_info = json.load(f)
+                # Prior to the following comparison, we need to adjust the value of "key" to
+                # be bytes, otherwise the comparison below will fail.
+                file_info["key"] = file_info["key"].encode()
+                if not self._equal_connections(info, file_info):
+                    raise ValueError(
+                        "Connection file already exists and does not match "
+                        "the expected values returned from provisioner!"
+                    )
+                file_exists = True
+
+        if not file_exists:
+            # Load the connection info and write out file. Note, this does not necessarily
+            # overwrite non-zero port values, so we'll validate afterward.
+            self.load_connection_info(info)
+            self.write_connection_file()
+
+        # Ensure what is in KernelManager is what we expect.  This will catch issues if the file
+        # already existed, yet it's contents differed from the KernelManager's (and provisioner).
+        km_info = self.get_connection_info()
+        if not self._equal_connections(info, km_info):
+            raise ValueError(
+                "KernelManager's connection information already exists and does not match "
+                "the expected values returned from provisioner!"
+            )
+
+    @staticmethod
+    def _equal_connections(conn1: KernelConnectionInfo, conn2: KernelConnectionInfo) -> bool:
+        """Compares pertinent keys of connection info data. Returns True if equivalent, False otherwise."""
+
+        pertinent_keys = [
+            "key",
+            "ip",
+            "stdin_port",
+            "iopub_port",
+            "shell_port",
+            "control_port",
+            "hb_port",
+            "transport",
+            "signature_scheme",
+        ]
+
+        for key in pertinent_keys:
+            if conn1.get(key) != conn2.get(key):
+                return False
+        return True
 
     # --------------------------------------------------------------------------
     # Creating connected sockets
@@ -657,7 +696,7 @@ class LocalPortCache(SingletonConfigurable):
     See: https://github.com/jupyter/jupyter_client/issues/487
     """
 
-    def __init__(self, **kwargs) -> None:
+    def __init__(self, **kwargs: Any) -> None:
         super().__init__(**kwargs)
         self.currently_used_ports: Set[int] = set()
 
